/******************************************************************************
 *
 *	Multi-Channel USB audio (SINICHIRO) library : sinich_lib.c
 *
 *	Copyright (c) 2006 by TED
 *
 *
 *	NOTES:
 *	 - 
 */


/******************************************************************************
 *	include files
 */
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdarg.h>
#include <signal.h>
#include <dlfcn.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <limits.h>

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#include "sinich_cmn.h"
#include "sinich_prm.h"
#include "sinich_lib.h"


/******************************************************************************
 *	definitions
 */
// debug print level
//   0 : print no message
//   1 : print major debug messages
//   2 : print detail debug messages
#define SINICH_LIB_DEBUG 0	










/******************************************************************************
 *	SINICHIRO driver open function
 *	accessing to audio board, I/O board, system board
 *
 *	parameter :
 *		io		: pointer pointer to SINICHIRO device handler
 *		name	: pointer to open device name
 *		dir		: stream direction (read/write)
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_sys_open(struct ich_io_t **io, uchar *name)
{
	int handle;

	// check arguments "name"
	if (NULL == name)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_open : ERROR : *name is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_sys_open : I/O handle ptr ptr 0x%08X, open device name %s, stream direction %d\n",
		   (int)io, name, dir);
#endif

	// SINICHIRO device open
//	handle = (int)open(name, 0x02);
	handle = (int)open(name, 0x02 | O_EXCL);

	// check SINICHIRO device open
	if (handle < 0)
//	if (0 > handle)
	{
		// error message
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_open : ERROR : cannot open device \"%s\"\n", name);
#endif
		return -1;
	}
	// I/O handle memory allocate
	*io = malloc(sizeof(struct ich_io_t));
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_open : ERROR : cannot allocate memory for I/O handle\n");
		return -1;
	}
	// set SINICHIRO device handle member
//	io->dir = dir;
	(*io)->dev = handle;

	return 0;
}


/******************************************************************************
 *	SINICHIRO driver close function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_sys_close(struct ich_io_t *io)
{
	int ret;

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_close : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_sys_close : I/O addr 0x%08X\n", (uint)io);
#endif

	// SINICHIRO device close
	ret = close(io->dev);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_close : ERROR : SINICHIRO close fail\n");
#endif
		return -1;
	}

	// free memory
	free(io);

	return 0;
}


/******************************************************************************
 *	set data to system board register function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *		reg		: system board register address
 *		ctl		: system board register data to set
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_sys_set_reg(struct ich_io_t *io, uchar reg, uchar ctl)
{
	int ret;
	uchar arg[2];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_set_reg : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_sys_set_reg : I/O addr 0x%08X, reg addr 0x%02X, set data 0x%02X\n",
		   (int)io, reg, ctl);
#endif

	// set register addr/data to temporary variable "arg"
	arg[0] = reg;
	arg[1] = ctl;

	// call I/O control
	// send "arg" address as "uint" data type
	//   arg[0] : system board register address
	//   arg[1] : system board register data to set
	ret = ioctl(io->dev, VR_SET_FPGA_DATA, (uint)arg);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_set_reg : ERROR : set register fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	get data from system board register function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *		reg		: system board register address
 *		ctl		: pointer for system board register data to get 
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_sys_get_reg(struct ich_io_t *io, uchar reg, uchar *ctl)
{
	int ret;
	uint arg[2];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_reg : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

	// check arguments "ctl"
	if (NULL == ctl)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_reg : ERROR : \"ctl\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_sys_get_reg : I/O addr 0x%08X, reg addr 0x%02X, ctl pointer 0x%08X\n",
		   (int)io, reg, (int)ctl);
#endif

	// set register addr/data to temporary variable "arg"
	arg[0] = (uint)reg;
	arg[1] = (uint)ctl;

	// call I/O control
	// send "arg" address as "uint" data type
	//   arg[0] : system board register address
	//   arg[1] : system board register data to set
	ret = ioctl(io->dev, VR_GET_FPGA_DATA, (uint)arg);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_reg : ERROR : get register fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	check stuck(Audio Codec) board ACT information data
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *		board	: stuck(Audio Codec) board number 
 *		info	: pointer for stuck board ACT register data
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_sys_act_info(struct ich_io_t *io, uchar board, uchar *info)
{
	int ret;
	uint arg[2];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_act_info : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

	// check arguments "info"
	if (NULL == info)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_act_info : ERROR : \"info\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_sys_act_info: I/O addr 0x%08X, board number 0x%02X, info pointer 0x%08X\n",
		   (uint)io, board, (uint)info);
#endif

	// set register addr/data to temporary variable "arg"
	arg[0] = (uint)board;
	arg[1] = (uint)info;

	// call I/O control
	// send "arg" address as "uint" data type
	//   arg[0] : stuck board number
	//   arg[1] : stuck board register data "Act check" pointer
	ret = ioctl(io->dev, SINICH_COM_SYS_ACT_INFO, (uint)arg);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_act_info : ERROR : get act info fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	SINICHIRO start function
 *	coordinate start timing, initialize structure ...
 *
 *	parameter :
 *		io		: pointer for SINICHIRO device handler
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_sys_start(struct ich_io_t *io)
{
	int ret;

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_start : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_sys_start : I/O addr 0x%08X\n", (int)io);
#endif

	// call I/O control
	ret = ioctl(io->dev, SINICH_COM_SYS_START, (uint)SINICH_PCM_START); //io->dir);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_start : ERROR : system start fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	SINICHIRO stop function
 *	coordinate stop timing, free memory ...
 *
 *	parameter :
 *		io		: pointer for SINICHIRO device handler
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_sys_stop(struct ich_io_t *io)
{
	int ret;

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_stop : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_sys_stop : I/O addr 0x%08X\n", (int)io);
#endif

	io->dir=SINICH_PCM_STOP;
	// call I/O control
	ret = ioctl(io->dev, SINICH_COM_SYS_STOP, (uint)io->dir);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_stop : ERROR : system stop fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}

// Jan 27, 2006 : add
/******************************************************************************
 *	sync start of all endpoint
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_all_start(struct ich_io_t *io)
{
	int ret;

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_all_start : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

	// call I/O control
	ret = ioctl(io->dev, VR_ALLEP_START, NULL);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_all_start : ERROR\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}

// Jan 27, 2006 : add
/******************************************************************************
 *	sync stop of all endpoint
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_all_stop(struct ich_io_t *io)
{
	int ret;

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_all_stop : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

	// call I/O control
	ret = ioctl(io->dev, VR_ALLEP_STOP, NULL);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_all_stop : ERROR\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}

/******************************************************************************
 *	get FIFO status function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *
 *	return value :
 *		ret		: success(status value) or error(-1)
 */
ushort sinich_sys_status(struct ich_io_t *io)
{
	int ret;

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_status : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_sys_status : I/O addr 0x%08X\n", (int)io);
#endif

	ret = ioctl(io->dev, SINICH_COM_SYS_STATUS, 0);
	if (ret < 0) {
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
	}
	return (ushort)ret;
}


#ifdef SINICH_STATUS_POLLING_WAIT_MODE_SUPPORT // not supported
/******************************************************************************
 *	wait unlock function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *
 *	return value :
 *		ret		: success(status value) or error(-1)
 */
unsigned long sinich_sys_ack(struct ich_io_t *io)
{
	int ret;

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_ack : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_sys_ack : I/O addr 0x%08X\n", (int)io);
#endif

	// call I/O control
	ret = ioctl(io->dev, SINICH_COM_SYS_ACK, (uint)NULL);

	return (unsigned long)ret;
}


/******************************************************************************
 *	wait I/O function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *		mode	: 0:polling/1:block
 *		timeout	: timeout value
 *
 *	return value :
 *		ret		: success(0) or timeout(0)
 */
int sinich_sys_wait(struct ich_io_t *io, uchar mode, uint timeout)
{
	int ret;
	uint arg[2];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_wait : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if(SINICH_LIB_DEBUG  > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_sys_wait : I/O addr 0x%08X, mode %d, timeout %d\n",
		   (uint)io, mode, timeout);
#endif

	// set register addr/data to temporary variable "arg"
	arg[0] = (uint)mode;
	arg[1] = timeout;

	// call I/O control
	// send "arg" address as "uint" data type
	//   arg[0] : mode
	//   arg[1] : timeout value
	ret = ioctl(io->dev, SINICH_COM_SYS_WAIT, (uint)arg);

	return ret;
}
#endif // #ifdef ICH_STATUS_POLLING_WAIT_MODE_SUPPORT // not supported


/******************************************************************************
 *	set sampling frequency and the number of channels function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *		ctl		: audio control register value
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_pcm_ctl(struct ich_io_t *io, uchar ctl)
{
	int ret;
	int arg[2];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_ctl : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if(SINICH_LIB_DEBUG  > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_pcm_ctl : I/O addr 0x%08X, ctl value 0x%02X\n",
		   (int)io, ctl);
#endif

	// call I/O control
	ret = ioctl(io->dev, SINICH_COM_PCM_CTL, (uint)ctl);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_ctl : ERROR : PCM control data set fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}	


/******************************************************************************
 *	set sampling frequency and the number of channels function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *		ctl		: audio control register value
 *
 *	return value :
 *		success(0) or error(-1)
 */
#if 0
int sinich_io_ctl(struct ich_io_t *io, uchar ctl)
{
	int ret;
	int arg[2];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_io_ctl : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if(SINICH_LIB_DEBUG  > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_io_ctl : I/O addr 0x%08X, ctl value 0x%02X\n",
		   (int)io, ctl);
#endif

	// call I/O control
	ret = ioctl(io->dev, SINICH_COM_IO_CTL, (uint)ctl);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_io_ctl : ERROR : I/O control data set fail\n");
#endif
		return -1;
	}

	return 0;
}	
#endif

/******************************************************************************
 *	set codec parameter function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *		board	: audio board number
 *		param	: pointer for codec parameter
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_pcm_set_codec(struct ich_io_t *io, uchar board, struct codec_param_t *param)
{
	int ret;
	uint arg[3];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_set_codec : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

	// check arguments "param"
	if (NULL == param)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_set_codec : ERROR : \"param\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_pcm_set_codec : I/O addr 0x%08X, board number 0x%02X, param pointer 0x%08X\n",
		   (int)io, board, (int)param);
#endif

	// set register addr/data to temporary variable "arg"
//	arg[0] = (uint)io->dir;
	arg[0] = (uint)0;
	arg[1] = (uint)board;
	arg[2] = (uint)param;

	// call I/O control
	// send "arg" address as "uint" data type
	//   arg[0] : stream
	//   arg[1] : audio board number
	//   arg[2] : audio board codec data pointer to set
	ret = ioctl(io->dev, SINICH_COM_PCM_SET_CODEC, (uint)arg);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_set_codec : ERROR : PCM CODEC set fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	get all codec parameter function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *		board	: audio board number ???
 *		param	: pointer for all codec parameter to get
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_pcm_get_codec(struct ich_io_t *io, uchar board, struct codec_param_t *param)
{
	int ret;
	uint arg[3];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_get_codec : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

	// check arguments "param"
	if (NULL == param)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_get_codec : ERROR : \"param\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_pcm_get_codec : I/O addr 0x%08X, board number 0x%02X, param pointer 0x%08X\n",
		   (int)io, board, (int)param);
#endif

	// set register addr/data to temporary variable "arg"
//	arg[0] = (uint)io->dir;
	arg[0] = (uint)0;
	arg[1] = (uint)board;
	arg[2] = (uint)param;


	// call I/O control
	// send "arg" address as "uint" data type
	//   arg[0] : stream
	//   arg[1] : audio board number
	//   arg[2] : audio board codec data pointer to get
	ret = ioctl(io->dev, SINICH_COM_PCM_GET_CODEC, (uint)arg);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_get_codec : ERROR : PCM CODEC get fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	PCM data write function
 *
 *	paremter :
 *		io		: pointer to SINICHIRO device handler
 *		buffer	:
 *		size	:
 *
 *	return value :
 *		ret		: success(written size) or error(-1)
 */
int sinich_pcm_write(struct ich_io_t *io, uchar *buffer, ulong size)
{
	int ret;
	uint arg[2];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_write : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

	// check arguments "buffer"
	if (NULL == buffer)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_write : ERROR : \"buffer\" is NULL pointer\n");
		return -1;
	}

	if(size == 0)	return 0;

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_pcm_write : I/O addr 0x%08X, buf addr 0x%08X, size %d\n",
		   (uint)io, (uint)buffer, size);
#endif

	// set buffer addr/size to temporary variable "arg"
	arg[0] = (uint)buffer;
	arg[1] = (uint)size;

	// call I/O control
	// send "arg" address as "uint" data type
	//	arg[0] : buffer address
	//	arg[1] : buffer size
	ret = ioctl(io->dev, SINICH_COM_PCM_BLKOUT, (uint)arg);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_write : ERROR : PCM write fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return ret;
}


/******************************************************************************
 *	PCM data write function
 *
 *	paremter :
 *		io		: pointer to SINICHIRO device handler
 *		buffer	:
 *		size	:
 *
 *	return value :
 *		ret		: success(written size) or error(-1)
 */
int sinich_pcm_write_dir(struct ich_io_t *io, uchar *buffer, ushort size)
{
	int ret;
	uint arg[2];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_write : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

	// check arguments "buffer"
	if (NULL == buffer)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_write : ERROR : \"buffer\" is NULL pointer\n");
		return -1;
	}

	if(size == 0)	return 0;

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_pcm_write : I/O addr 0x%08X, buf addr 0x%08X, size %d\n",
		   (uint)io, (uint)buffer, size);
#endif

	// set buffer addr/size to temporary variable "arg"
	arg[0] = (uint)buffer;
	arg[1] = (uint)size;

	// call I/O control
	// send "arg" address as "uint" data type
	//	arg[0] : buffer address
	//	arg[1] : buffer size
	ret = ioctl(io->dev, SINICH_COM_PCM_WRITE, (uint)arg);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_write : ERROR : PCM write fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return ret;
}


/******************************************************************************
 *	PCM data read function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *		buffer	:
 *		size	:
 *
 *	return value :
 *		ret		: success(read size) or error(-1)
 */
int sinich_pcm_read(struct ich_io_t *io, uchar *buffer, ulong size)
{
	int ret;
	uint arg[2];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_read : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

	// check arguments "buffer"
	if (NULL == buffer)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_read : ERROR : \"buffer\" is NULL pointer\n");
		return -1;
	}

	if(size == 0)	return 0;

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_pcm_read : I/O addr 0x%08X, buf addr 0x%08X, size %d\n",
		   (int)io, (int)buffer, size);
#endif

	// set buffer addr/size data to temporary variable "arg"
	arg[0] = (uint)buffer;
	arg[1] = (uint)size;

	// call I/O control
	// send "arg" address as "uint" data type
	//   arg[0] : buffer address
	//   arg[1] : buffer size
	ret = ioctl(io->dev, SINICH_COM_PCM_BLKIN, (uint)arg);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_read : ERROR : PCM read fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return ret;
}


/******************************************************************************
 *	PCM data read function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *		buffer	:
 *		size	:
 *
 *	return value :
 *		ret		: success(read size) or error(-1)
 */
int sinich_pcm_read_dir(struct ich_io_t *io, uchar *buffer, ushort size)
{
	int ret;
	uint arg[2];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_read : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

	// check arguments "buffer"
	if (NULL == buffer)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_read : ERROR : \"buffer\" is NULL pointer\n");
		return -1;
	}

	if(size == 0)	return 0;

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_pcm_read : I/O addr 0x%08X, buf addr 0x%08X, size %d\n",
		   (int)io, (int)buffer, size);
#endif

	// set buffer addr/size data to temporary variable "arg"
	arg[0] = (uint)buffer;
	arg[1] = (uint)size;

	// call I/O control
	// send "arg" address as "uint" data type
	//   arg[0] : buffer address
	//   arg[1] : buffer size
	ret = ioctl(io->dev, SINICH_COM_PCM_READ, (uint)arg);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_pcm_read : ERROR : PCM read fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return ret;
}


/******************************************************************************
 *	get CODEC(audio, I/O) control data function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *		board	: the board number (0-3)
 *		reg		: register address
 *		data	: regstier data pointer
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_sys_get_codec_ctl(struct ich_io_t *io, uchar board_sel, uchar codec_sel, uchar codec_addr, uchar *codec_data)
{
	int ret;
	uint arg[4];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_codec_ctl : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

	// check arguments "codec_data"
	if (NULL == codec_data)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_codec_ctl : ERROR : \"codec_data\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_codec_ctl: I/O ptr 0x%08X, board number 0x%02X, codec chip number 0x%02X\n",
		   (uint)io, board_sel, codec_sel);
	printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_codec_ctl: codec control address 0x%02X, codec control data ptr 0x%08X\n",
		   codec_addr, (uint)codec_data);
#endif

	// set buffer addr/size data to temporary variable "arg"
	arg[0] = (uint)board_sel;
	arg[1] = (uint)codec_sel;
	arg[2] = (uint)codec_addr;
	arg[3] = (uint)codec_data;

	// call I/O control
	// send "arg" address as "uint" data type
	//   arg[0] : board select number : 0-7
	//   arg[1] : codec select number : 0-3
	//   arg[2] : codec control register address
	//   arg[3] : codec control register data
	ret = ioctl(io->dev, SINICH_COM_SYS_GET_CODEC_CTL, (uint)arg);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_codec_ctl : ERROR : audio or I/O control data get fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	get CODEC(audio, I/O) control data function
 *
 *	parameter :
 *		io			: pointer to SINICHIRO device handler
 *		board_sel	: board select number : 0-3
 *		codec_sel	: codec select number : 0-3
 *		codec_addr	: codec control register address
 *		codec_data	: codec control regstier data
 *
 *	return value :
 *		success(0) or error(-1)
 */
int sinich_sys_set_codec_ctl(struct ich_io_t *io, uchar board_sel, uchar codec_sel, uchar codec_addr, uchar codec_data)
{
	int ret;
	uchar arg[4];

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_set_codec_ctl : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_sys_set_codec_ctl: I/O ptr 0x%08X, board number 0x%02X, codec chip number 0x%02X\n",
		   (uint)io, board_sel, codec_sel);
	printf("SINICH-LIB : sinich_lib.c : sinich_sys_set_codec_ctl: codec control address 0x%02X, codec control data 0x%02X\n",
		   codec_addr, codec_data);
#endif

	// set buffer addr/size data to temporary variable "arg"
	arg[0] = board_sel;
	arg[1] = codec_sel;
	arg[2] = codec_addr;
	arg[3] = codec_data;

	// call I/O control
	// send "arg" address as "int" data type
	//   arg[0] : board select number : 0-7
	//   arg[1] : codec select number : 0-3
	//   arg[2] : codec control register address
	//   arg[3] : codec control register data
	ret = ioctl(io->dev, SINICH_COM_SYS_SET_CODEC_CTL, (uint)arg);

	// check return value
	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_set_codec_ctl : ERROR : audio or I/O control data set fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}


/******************************************************************************
 *	FPGA version number get function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *
 *	return value :
 *		version	: success(version number) or error(-1)
 */
int sinich_sys_get_fpga_version(struct ich_io_t *io)
{
	int version;
	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_fpga_version : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_fpga_version: I/O addr 0x%08X\n", (uint)io);
#endif

	// call I/O control
	version = ioctl(io->dev, SINICH_COM_SYS_GET_FPGA_VERSION, (uint)NULL);

	// check return value
	if (version < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_fpga_version : ERROR : FPGA version number get fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return version;
}


/******************************************************************************
 *	device driver version number get function
 *
 *	parameter :
 *		io		: pointer to SINICHIRO device handler
 *
 *	return value :
 *		version	: success(version number) or error(-1)
 */
int sinich_sys_get_driver_version(struct ich_io_t *io)
{
	int version;

	// check arguments "io"
	if (NULL == io)
	{
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_driver_version : ERROR : \"io\" is NULL pointer\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_driver_version\n");
#endif

	// call I/O control
	version = ioctl(io->dev, SINICH_COM_SYS_GET_DRIVER_VERSION, (uint)NULL);

	// check return value
	if (version < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : sinich_sys_get_driver_version : ERROR : driver version number get fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return version;
}


/******************************************************************************
 *	USB device descriptor get function
 *
 *	parameter :
 *		hdev	: device handle
 *		dsc		: USB device descriptor pointer
 *
 *	return value :
 *		ret		: success(0) or error(-1)
 */
int get_device_descriptor(HANDLE hdev, PUSB_DEVICE_DESCRIPTOR dsc)
{
	int ret;
	uint arg[1];

	// check arguments "io"
	if ((PUSB_DEVICE_DESCRIPTOR)NULL == dsc)
	{
		printf("SINICH-LIB : sinich_lib.c : get_device_descriptor : ERROR : \"dsc\" is NULL\n");
		return -1;
	}

#if (SINICH_LIB_DEBUG > 0)
	printf("SINICH-LIB : sinich_lib.c : get_device_descriptor : hdev 0x%02X, dsc 0x%08X\n", hdev, (uint)dsc);
#endif

	// set buffer addr/size data to temporary variable "arg"
	arg[0] = (uint)dsc;

	// call I/O control
	// send "arg" address as "uint" data type
	//   arg[0] : USB device descriptor pointer
	ret = ioctl(hdev, SINICH_COM_SYS_GET_USB_DEVICE_DESCRIPTOR, (uint)arg);

	if (ret < 0)
	{
#if (SINICH_LIB_DEBUG > 1)
		printf("SINICH-LIB : sinich_lib.c : get_device_descriptor : ERROR : I/O control operation fail\n");
#endif
		printf("%s %d : %s\n", __FILE__, __LINE__, strerror(errno));
		return -1;
	}

	return 0;
}

